"""Scopus Client module.

This module is responsible to provide an efficient, and error proof client for the Scopus API.
The client is optimized to scrape the maximum number of pages that are available from Scopus API,
which is 5000, for normal users, at the maximum speed possible.

We use [`aiometer`](https://github.com/florimondmanca/aiometer),
and [`httpx`](https://github.com/projectdiscovery/httpx) to achieve this goal.
"""  # noqa: E501

from dataclasses import dataclass
from datetime import datetime
from typing import AsyncIterator, List, Optional, Union

from more_itertools import divide

from . import api


class OutOfAPIKeysError(Exception):
    """Used all API keys available."""


class ExceededTimeoutRetriesError(Exception):
    """Exceeded the number of timeout retries in a row."""


@dataclass
class APIKeyExpiredResponse:
    """Represents an API key expired response from the [`ScopusClient`][sesg.scopus.client.ScopusClient].

    Args:
        api_key (str): The expired API key.
        api_key_index (int): The index of the expired API key on the list of API keys passed to [`ScopusClient`][sesg.scopus.client.ScopusClient].
        resets_at (Optional[datetime]): Datetime object represents when the API key will be resetted.
    """  # noqa: E501

    api_key: str
    api_key_index: int
    resets_at: Optional[datetime]


@dataclass
class TimeoutResponse:
    """Represents a timed out response from the [`ScopusClient`][sesg.scopus.client.ScopusClient].

    Args:
        timed_out_page (int): The page where the timeout occured.
        timeout_retry (int): The current timeout retry attempt for a same request.
    """  # noqa: E501

    timed_out_page: int
    timeout_retry: int


class ScopusClient:
    """A Scopus API Client with automatic retry on timeout, and automatic API key swapping on expiry."""  # noqa: E501

    __timeout: float
    __api_keys: List[str]
    __timeout_retries: int

    __current_api_key_index: int
    __current_timeout_retry: int
    __current_page: int
    __current_query: str

    def __init__(
        self,
        *,
        api_keys: List[str],
        timeout: float,
        timeout_retries: int,
    ) -> None:
        """Creates a ScopusClient.

        Args:
            api_keys (List[str]): List with Scopus API keys.
            timeout (float): Time in seconds to wait before assuming the request timed out.
            timeout_retries (int): Number of times to retry a timed out request in a row.
        """  # noqa: E501
        self.__timeout = timeout
        self.__api_keys = api_keys
        self.__timeout_retries = timeout_retries

        self.__current_api_key_index = 0
        self.__current_timeout_retry = 0
        self.__current_page = 0

    @property
    def current_api_key(self) -> str:
        """Current API key being used."""
        return self.__api_keys[self.__current_api_key_index]

    @property
    def current_api_key_index(self) -> int:
        """Index on the list of the current API key being used."""
        return self.__current_api_key_index

    @property
    def current_page(self) -> int:
        """Current page being fetched.

        Starts at 1, being at most 200.
        """
        return self.__current_page

    @property
    def current_query(self) -> str:
        """Current query being used."""
        return self.__current_query

    @property
    def api_keys(self) -> List[str]:
        """The Scopus API keys for this client."""
        return self.__api_keys

    @property
    def current_timeout_retry(self) -> int:
        """The current retry for the same timed out request."""
        return self.__current_timeout_retry

    def __restart_iterator_with_current_state(self):
        return api.search(
            api_key=self.current_api_key,
            query=self.__current_query,
            timeout=self.__timeout,
            page=self.__current_page,
        )

    async def __anext__(
        self,
    ) -> Union[api.SuccessResponse, APIKeyExpiredResponse, TimeoutResponse,]:
        """Returns the next result of the search iterator.

        Raises:
            ExceededTimeoutRetriesError: If exceeds the number of timeout retries in a row.
            OutOfAPIKeysError: If used al API keys available.
            PayloadTooLargeError: If the response status code is 413. Probably indicates that the search string is too large.
            BadRequestError: If the response status code is 400. Probably indicates that the search string is malformed.

        Returns:
            One of the below:

                - A succesfull response
                - An API key expired response
                - A Timed out response
        """  # noqa: E501
        try:
            response = await self.__iterator.__anext__()

            self.__current_page += 1
            self.__current_timeout_retry = 0

            return response

        except (api.PayloadTooLargeError, api.BadRequestError) as e:
            raise e

        except api.APIKeyExpiredError as e:
            response = APIKeyExpiredResponse(
                api_key=self.current_api_key,
                api_key_index=self.current_api_key_index,
                resets_at=e.resets_at,
            )

            self.__current_timeout_retry = 0
            self.__current_api_key_index += 1
            self.__iterator = self.__restart_iterator_with_current_state()

            if self.__current_api_key_index == len(self.__api_keys):
                raise OutOfAPIKeysError()

            return response

        except api.TimeoutError:
            if self.__current_timeout_retry == self.__timeout_retries - 1:
                raise ExceededTimeoutRetriesError()

            response = TimeoutResponse(
                timed_out_page=self.current_page,
                timeout_retry=self.current_timeout_retry,
            )

            self.__current_timeout_retry += 1
            self.__iterator = self.__restart_iterator_with_current_state()

            return response

    def __aiter__(self):
        """Returns an async iterator over the results of the current query."""
        self.__current_page = 0
        self.__current_timeout_retry = 0
        self.__iterator = self.__restart_iterator_with_current_state()

        return self

    def search(
        self,
        query: str,
    ) -> AsyncIterator[
        Union[
            api.SuccessResponse,
            APIKeyExpiredResponse,
            TimeoutResponse,
        ]
    ]:
        """Performs Scopus API requests in a timeout-proof, and API key expiry-prof manner.

        !!! note

            Notice that the exceptions on the **Raises**
            section occurs during iteration time, not at function call time.

        Args:
            query (str): The query to search for.

        Raises:
            ExceededTimeoutRetriesError: If exceeds the number of timeout retries in a row.
            OutOfAPIKeysError: If used al API keys available.
            PayloadTooLargeError: If the response status code is 413. Probably indicates that the search string is too large.
            BadRequestError: If the response status code is 400. Probably indicates that the search string is malformed.

        Returns:
            An AsyncIterator that yields one of the below:

                - A succesfull response
                - An API key expired response
                - A Timed out response

        Examples:
            >>> client = ScopusClient(api_keys=[], timeout=7, timeout_retries=10)
            >>> async for data in client.search("machine"):  # doctest: +SKIP
            ...     if isinstance(data, TimeoutResponse):
            ...         print(f"Timed out on page {client.current_page}")
            ...
            ...     elif isinstance(data, APIKeyExpiredResponse):
            ...         print(f"API Key {client.current_api_key_index + 1} of {len(client.api_keys)} is expired.")
            ...
            ...     else:
            ...         # here data is of type `SuccessResponse`
            ...         print(data.entries)
        """  # noqa: E501
        self.__current_query = query

        return self.__aiter__()


def create_clients_with_disjoint_api_keys(
    *,
    api_keys: List[str],
    timeout: float,
    timeout_retries: int,
    n_clients: int,
) -> List[ScopusClient]:
    """Creates `n_clients` ScopusClient instances, where each client will use a disjoint set API keys.

    If the number of API keys is not divisible by `n_clients`, some clients will have fewer API keys than the others.

    Args:
        api_keys (List[str]): List with Scopus API keys.
        timeout (float): Time in seconds to wait before assuming the request timed out.
        timeout_retries (int): Number of times to retry a timed out request in a row.
        n_clients (int): Number of clients to create

    Raises:
        ValueError: If the number of clients is greater than the number of API keys.

    Returns:
        A list with the created clients.
    """  # noqa: E501
    if n_clients > len(api_keys):
        # fmt: off
        raise ValueError("Number of clients must be smaller than the number of API keys.")  # noqa: E501

    api_keys_groups = divide(n_clients, api_keys)

    clients_list: List[ScopusClient] = list()

    for group in api_keys_groups:
        client = ScopusClient(
            api_keys=list(group),
            timeout=timeout,
            timeout_retries=timeout_retries,
        )

        clients_list.append(client)

    return clients_list



"""Snowballing module.

This module is responsible to perform the snowballing method on a set of studies.
In details, given a set of studies with titles, and text contents, for each study,
we try to find out whether the study cited another given study or not.

To achieve this goal, we use
[`rapidfuzz`](https://github.com/maxbachmann/RapidFuzz)
to perform string similarity checks.
"""  # noqa: E501

from itertools import islice
from multiprocessing import Pool
from typing import Iterable, Iterator, List, Tuple, TypedDict, TypeVar

from rapidfuzz import process


T = TypeVar("T")


def _window(
    *,
    seq: Iterable[T],
    size: int,
) -> Iterator[Tuple[T, ...]]:
    """Creates an iterator over overlapping subslices of the given size.

    Args:
        seq (Iterable[T]): Sequence to iterate over.
        size (int): Size of each subslice.

    Yields:
        A subslice with the given size.

    Examples:
        >>> elements = [1, 2, 3, 4, 5, 6]
        >>> for subslice in _window(seq=elements, size=3):
        ...     print(subslice)
        (1, 2, 3)
        (2, 3, 4)
        (3, 4, 5)
        (4, 5, 6)
    """
    it = iter(seq)
    result = tuple(islice(it, size))

    if len(result) == size:
        yield result

    for elem in it:
        result = result[1:] + (elem,)
        yield result


def _study_cites_title(
    *,
    study: str,
    title: str,
) -> bool:
    """Uses `thefuzz.process.extractOne` to determine if a study cites a title.

    Args:
        study (str): Text of the study.
        title (str): Title to search for.

    Returns:
        True if the study cites the title, False otherwise.

    Examples:
        >>> _study_cites_title(
        ...     title="regression tests for machine learning models: a systematic literature review",
        ...     study="long text here very long REFERENCES: regression tests for machine learning models: a systematic literature review",
        ... )
        True
    """  # noqa: E501
    window_size = len(title)
    options = ["".join(x) for x in _window(seq=study, size=window_size)]

    result = process.extractOne(title, options)

    if result is not None and result[1] >= 90:
        return True

    return False


class _PooledStudyCitesTitleArgs(TypedDict):
    """Data container for the `_pooled_study_cites_title` function, with the goal of improving type safety.

    Args:
        title (str): Title to search for
        study (str): Text of the study.
        skip (bool): Indicates if should skip the execution and return False.
    """  # noqa: E501

    title: str
    study: str
    skip: bool


def _pooled_study_cites_title(
    args: _PooledStudyCitesTitleArgs,
) -> bool:
    """Replicates `_study_cites_title` behaviour, with slight modifications to work well with `multiprocessing.Pool`.

    Args:
        args (_PooledStudyCitesTitleArgs): args of this function.

    Returns:
        False if skip is True, the result of `_study_cites_title(args["title"], args["study"])` otherwise.

    Examples:
        >>> _pooled_study_cites_title(
        ...     _PooledStudyCitesTitleArgs(
        ...         title="regression tests for machine learning models: a systematic literature review",
        ...         study="TITLE: regression tests for machine learning models: a systematic literature review. Abstract: abstract here",
        ...         skip=True,
        ...     )
        ... )
        False
    """  # noqa: E501
    if args["skip"]:
        return False

    return _study_cites_title(study=args["study"], title=args["title"])


def _preprocess_title(
    *,
    title: str,
) -> str:
    """Processes the title in the following manner.

    1. Removes leading and trailing whitespaces
    1. Turns to lower case
    1. Removes spaces and dots

    Args:
        title (str): Title to preprocess.

    Returns:
        Preprocessed title.

    Examples:
        >>> _preprocess_title(title=" title. HERE ")
        'titlehere'
    """
    return title.strip().lower().replace(" ", "").replace(".", "")


def _preprocess_study(
    *,
    study_text: str,
) -> str:
    r"""Processes the study in the following manner.

    1. Removes leading and trailing whitespaces
    1. Turns to lower case
    1. Removes line breaks, line carriages, spaces, and dots

    Args:
        study_text (str): Study's text to preprocess

    Returns:
        Preprocessed study.

    Examples:
        >>> _preprocess_study(study_text=" text. \n \r\n HERE ")
        'texthere'
    """
    return (
        study_text.strip()
        .lower()
        .replace("\n", "")
        .replace("\r", "")
        .replace(" ", "")
        .replace(".", "")
    )


class SnowballingStudy:
    r"""Represents a study that will be included in backward snowballing.

    The constructor will preprocess the title and text content to the correct format.

    Examples:
        >>> s = SnowballingStudy(id=1, title=" title. HERE ", text_content=" text. \n \r\n HERE ")
        >>> s.title == "titlehere", s.text_content == "texthere"
        (True, True)
    """  # noqa: E501

    __id: int
    __title: str
    __text_content: str

    def __init__(
        self,
        *,
        id: int,
        title: str,
        text_content: str,
    ) -> None:
        """Creates an instance of a SnowballingStudy.

        Args:
            id (int): Identifier of the study. Could be a database id, for example.
            title (str): Title of the study.
            text_content (str): Content of the study. Could be extracted from a PDF with CERMINE.
        """  # noqa: E501
        self.__id = id
        self.__title = _preprocess_title(title=title)
        self.__text_content = _preprocess_study(study_text=text_content)

    @property
    def id(self) -> int:
        """ID of the study."""
        return self.__id

    @property
    def title(self) -> str:
        """Title of the study."""
        return self.__title

    @property
    def text_content(self) -> str:
        """Text content of the study."""
        return self.__text_content


def backward_snowballing(
    *,
    studies: List[SnowballingStudy],
) -> Iterator[Tuple[SnowballingStudy, List[SnowballingStudy]]]:
    """Runs backward snowballing in the given list of studies.

    Args:
        studies (List[SnowballingStudy]): List of studies with id, title, and text content.

    Yields:
        Iterator of tuples, where the tuple holds a study, and the studies that are referenced.

    Examples:
        >>> studies: List[SnowballingStudy] = [
        ...     SnowballingStudy(id=1, title="title 1", text_content="... REFERENCES: machine learning, a SLR"),
        ...     SnowballingStudy(id=2, title="machine learning, a SLR", text_content="... REFERENCES: other studies"),
        ... ]
        >>>
        >>> for study, references in backward_snowballing(studies=studies):
        ...     print((study.id, [r.id for r in references]))
        (1, [2])
        (2, [])
    """  # noqa: E501
    for study_index, study in enumerate(studies):
        with Pool() as p:
            # if `result[j]` is True then the current study cites title `j`
            result: List[bool] = p.map(
                _pooled_study_cites_title,
                [
                    {
                        # when `i == j`, we are checking if study `i` cites itself,
                        # so we skip and set it as False
                        "skip": study_index == reference_index,
                        "study": study.text_content,
                        "title": reference.title,
                    }
                    for reference_index, reference in enumerate(studies)
                ],
            )

        references: List[SnowballingStudy] = list()

        for result_index, is_cited in enumerate(result):
            if is_cited:
                referenced_study = studies[result_index]
                references.append(referenced_study)

        yield study, references



"""Metrics module.

This module is responsible to provide metrics of a search string
performance, given the number of Scopus results found, number of GS
studies found in the Scopus results, and other informations.
"""  # noqa: E501

from dataclasses import dataclass
from typing import Any, List, Tuple

from rapidfuzz.distance import Levenshtein


@dataclass
class Metrics:
    """Data container for metrics calculations.

    Args:
        gs_size (int): Number of studies in the GS.
        n_scopus_results (int): Total number of Scopus results. Don't set the 5_000 boundary, as this class will automatically use the correct value when calculating the metrics.
        n_qgs_studies_in_scopus (int): Number of QGS studies that were found in Scopus.
        n_gs_studies_in_scopus (int): Number of GS studies that were found in Scopus.
        n_gs_studies_in_scopus_and_bsb (int): Number of GS studies that were found in Scopus, and can be found via backward snowballing on the start set.
        n_gs_studies_in_scopus_and_bsb_and_fsb (int): Number of GS studies that were found in Scopus, and can be found via backward or forward snowballing on the start set.

    Examples:
        >>> metrics = Metrics(
        ...     gs_size=15,
        ...     n_scopus_results=13,
        ...     n_qgs_studies_in_scopus=2,
        ...     n_gs_studies_in_scopus=2,
        ...     n_gs_studies_in_scopus_and_bsb=8,
        ...     n_gs_studies_in_scopus_and_bsb_and_fsb=15,
        ... )
        >>> metrics.scopus_and_bsb_and_fsb_recall
        1.0
    """  # noqa: E501

    gs_size: int
    n_scopus_results: int

    n_qgs_studies_in_scopus: int
    n_gs_studies_in_scopus: int
    n_gs_studies_in_scopus_and_bsb: int
    n_gs_studies_in_scopus_and_bsb_and_fsb: int

    @property
    def n_scopus_results_with_limit(self):
        """Limits the number of results up to 5000."""
        return min(5000, self.n_scopus_results)

    @property
    def scopus_precision(self):
        """Measures the precision using the number of GS studies found in Scopus, and the number of Scopus Results with an upper limit of 5000.

        If no Scopus result were found, will return 0.
        """  # noqa: E501
        if self.n_scopus_results_with_limit == 0:
            return 0

        return self.n_gs_studies_in_scopus / self.n_scopus_results_with_limit

    @property
    def scopus_recall(self):
        """Measures the recall using the number of GS studies found in Scopus, and the GS size."""  # noqa: E501
        return self.n_gs_studies_in_scopus / self.gs_size

    @property
    def scopus_f1_score(self):
        """Measures the F1-Score using `scopus_precision` and `scopus_recall`.

        If the recall and the precision are both 0 (which happens when
        no GS studies were found in Scopus results, and when no Scopus
        results were found, respectively), will return 0.
        """
        precision = self.scopus_precision
        recall = self.scopus_recall

        numerator = 2 * precision * recall
        denominator = precision + recall

        if denominator == 0:
            return 0

        return numerator / denominator

    @property
    def scopus_and_bsb_recall(self):
        """Measures the recall using the number of GS studies found in Scopus + number of studies found with Backward Snowballing, and the GS size."""  # noqa: E501
        return self.n_gs_studies_in_scopus_and_bsb / self.gs_size

    @property
    def scopus_and_bsb_and_fsb_recall(self):
        """Measures the recall using the number of GS studies found in Scopus + number of studies found with Backward Snowballing, + number of studiesfound with Forward Snowballing, and the GS size."""  # noqa: E501
        return self.n_gs_studies_in_scopus_and_bsb_and_fsb / self.gs_size


def preprocess_string(
    *,
    string: str,
) -> str:
    r"""Strips the string and turn every character to lower case.

    Args:
        string: The string to preprocess.

    Returns:
        The preprocessed string.

    Examples:
        >>> preprocess_string(string=" A string Here.  \n")
        'a string here.'
    """
    return string.strip().lower()


def similarity_score(
    *,
    small_set: List[str],
    other_set: List[str],
) -> List[Tuple[int, int]]:
    """Uses `TfidfVectorizer`, `cosine_similarity`, and `Levenshtein` to calculate the intersection of two sets of strings.

    You might need to preprocess the strings with [sesg.metrics.preprocess_string][].

    Args:
        small_set (List[str]): Set of strings. If possible, the length of this set should be smaller than the other one.
        other_set (List[str]): Set of strings to compare against.

    Returns:
        List of tuples, where the tuple `(i, j)` means that `small_set[i]` is similar to `other_set[j]`.

    Examples:
        >>> small_set = ["machine learning", "databases", "search strings"]
        >>> other_set = ["Databases, an introduction", "Machine Learning", "Search String"]
        >>> similarity_score(
        ...     small_set=[preprocess_string(string=s) for s in small_set],
        ...     other_set=[preprocess_string(string=s) for s in other_set]
        ... )
        [(0, 1), (2, 2)]
    """  # noqa: E501
    from numpy import argsort
    from sklearn.feature_extraction.text import TfidfVectorizer
    from sklearn.metrics.pairwise import cosine_similarity

    # without this "Any typings", pylance takes too long to analyze the sklearn files
    # remove this line once sklearn has developed stubs for the package
    TfidfVectorizer: Any
    cosine_similarity: Any

    if len(other_set) == 0:
        return []

    train_set = [*small_set, *other_set]

    tfidf_vectorizer = TfidfVectorizer()
    tfidf_matrix = tfidf_vectorizer.fit_transform(train_set)

    first_set_matrix = tfidf_matrix[0 : len(small_set)]
    second_set_matrix = tfidf_matrix[len(small_set) : len(small_set) + len(other_set)]

    similarity_matrix = cosine_similarity(
        first_set_matrix,
        second_set_matrix,
    )

    lines: int
    lines, _ = similarity_matrix.shape

    similars: List[Tuple[int, int]] = []

    for index_of_first_set_element in range(lines):
        # contains the row of the similarity matrix for the current element
        line = similarity_matrix[index_of_first_set_element]

        index_of_closest_element_in_second_set: int = argsort(line)[-1]

        first_set_element = small_set[index_of_first_set_element]
        second_set_element = other_set[index_of_closest_element_in_second_set]

        distance = Levenshtein.distance(
            first_set_element,
            second_set_element,
            score_cutoff=10,
        )

        if distance < 10:
            similars.append(
                (index_of_first_set_element, index_of_closest_element_in_second_set)
            )

    return similars
